#include "glob-search.h"
#include "glob-word.h"

enum
{
    FINISHED,
    LAST_SIGNAL
};

struct _GlobSearch
{
    GObject parent_instance;

    gchar **wlist;
    guint length;
};

static GlobSearch *singleton = NULL;

G_DEFINE_TYPE (GlobSearch, glob_search, G_TYPE_OBJECT)

static guint signals[LAST_SIGNAL] = { 0, };

static void
glob_search_finalize (GObject *object)
{
    GlobSearch *search = GLOB_SEARCH (object);

    g_strfreev (search->wlist);

    G_OBJECT_CLASS (glob_search_parent_class)->finalize (object);
}

static GObject *
glob_search_constructor (GType                  type,
                         guint                  n_construct_params,
                         GObjectConstructParam *construct_params)
{
    GObject *ret;

    if (singleton)
        return g_object_ref (singleton);

    ret = G_OBJECT_CLASS (glob_search_parent_class)->constructor (type,
                                                                  n_construct_params,
                                                                  construct_params);

    singleton = GLOB_SEARCH (ret);
    g_object_add_weak_pointer (ret, (gpointer) &singleton);

    return ret;
}

static void
glob_search_class_init (GlobSearchClass *klass)
{
    GObjectClass *obj_class = G_OBJECT_CLASS (klass);

    obj_class->finalize = glob_search_finalize;
    obj_class->constructor = glob_search_constructor;

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1, G_TYPE_UINT);
}

static void
load_contents_cb (GObject      *object,
                  GAsyncResult *res,
                  gpointer      user_data)
{
    GlobSearch *search = GLOB_SEARCH (user_data);
    GError *err = NULL;
    gchar *contents;

    if (!g_file_load_contents_finish (G_FILE (object), res, &contents, NULL, NULL, &err))
        g_error ("Error: %s", err->message);

    search->wlist = g_strsplit (contents, "\n", 0);
    search->length = g_strv_length (search->wlist);

    g_free (contents);
}

static void
glob_search_init (GlobSearch *search)
{
    GFile *file;

    file = g_file_new_for_path ("wlist");
    g_file_load_contents_async (file, NULL, load_contents_cb, search);
}

/**
 * glob_search_dup:
 *
 * Duplicates the default #GlobSearch. This is used to search the word list.
 *
 * Returns: the default #GlobSearch
 */
GlobSearch *
glob_search_dup (void)
{
    return g_object_new (GLOB_TYPE_SEARCH, NULL);
}

/**
 * glob_search_get_matches:
 * @search: the search object
 * @glob: the pattern to match against
 *
 * Gets the list of words that @glob matches against.
 *
 * Returns: (transfer full) the list of words, free with g_strfreev() when finished
 */
static GListModel *
glob_search_get_matches (GlobSearch  *search,
                         const gchar *glob)
{
    GPatternSpec *pattern;
    GListStore *store;
    guint i;

    pattern = g_pattern_spec_new (glob);
    store = g_list_store_new (GLOB_TYPE_WORD);

    for (i = 0; i < search->length; i++)
    {
        if (g_pattern_match_string (pattern, search->wlist[i]))
        {
            GlobWord *word = glob_word_new (search->wlist[i]);
            g_list_store_append (store, word);
        }
    }

    g_pattern_spec_free (pattern);

    guint n = g_list_model_get_n_items (G_LIST_MODEL (store));
    g_signal_emit (search, signals[FINISHED], 0, n);
    return G_LIST_MODEL (store);
}

static void
get_matches_async_thread (GTask *task,
                          gpointer object,
                          gpointer task_data,
                          GCancellable *cancellable)
{
    const gchar *glob;
    GListModel *matches;

    glob = (const gchar *) task_data;
    matches = glob_search_get_matches (GLOB_SEARCH (object), glob);
    g_task_return_pointer (task, matches, (GDestroyNotify) g_object_unref);
}

/**
 * glob_search_get_matches_async:
 * @search: a #GlobSearch
 * @glob: the pattern to match against
 * @cancellable: (allow-none): optional #GCancellable object, %NULL to ignore
 * @callback: (scope async): the callback to run when finished
 * @user_data: (closure): data to pass to the callback
 *
 * Gets the list of words that @glob matches against asynchronously.
 * When this finishes, @callback will be called. Call
 * glob_search_get_matches_finish() to get the result.
 */
void
glob_search_get_matches_async (GlobSearch          *search,
                               const gchar         *glob,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  callback,
                               gpointer             user_data)
{
    GTask *task;

    g_return_if_fail (GLOB_IS_SEARCH (search));
    g_return_if_fail (glob != NULL);

    task = g_task_new (search, cancellable, callback, user_data);
    g_task_set_task_data (task, (gpointer) glob, NULL);
    g_task_run_in_thread (task, get_matches_async_thread);
    g_object_unref (task);
}

/**
 * glob_search_get_matches_finish:
 * @search: a #GlobSearch.
 * @result: a #GAsyncResult
 * @error: a #GError, %NULL to ignore
 * 
 * Finishes an asynchronous pattern matching operation.
 * 
 * Returns: (transfer full): the list of words, free with g_object_unref() when
 *     finished
 */
GListModel *
glob_search_get_matches_finish (GlobSearch    *search,
                                GAsyncResult  *res,
                                GError       **error)
{
    g_return_val_if_fail (GLOB_IS_SEARCH (search), NULL);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (res), NULL);
    g_return_val_if_fail (g_task_is_valid (res, search), NULL);

    return g_task_propagate_pointer (G_TASK (res), error);
}
